"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

storage.k8s.io/v1/StorageClass
"""
import copy
from typing import Any
from typing import Dict
from typing import List

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    provisioner: str,
    allow_volume_expansion: bool = True,
    allowed_topologies: List = None,
    mount_options: List = None,
    parameters: Dict = None,
    reclaim_policy: str = None,
    volume_binding_mode: str = "Immediate",
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Create a StorageClass.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        allow_volume_expansion(bool, optional): AllowVolumeExpansion shows whether the storage class allow volume
            expand.
        allowed_topologies(List, optional): Restrict the node topologies where volumes can be dynamically provisioned.
            Each volume plugin defines its own supported topology specifications.
            An empty TopologySelectorTerm list means there is no topology restriction.
            This field is only honored by servers that enable the VolumeScheduling feature.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        mount_options(List, optional): Dynamically provisioned PersistentVolumes of this storage class are created with
            these mountOptions, e.g. ["ro", "soft"]. Not validated - mount of the PVs will simply fail if one is
            invalid.
        parameters(Dict, optional): Parameters holds the parameters for the provisioner that should create volumes of
            this storage class.
        provisioner(Text): Provisioner indicates the type of the provisioner.
        reclaim_policy(Text, optional): Dynamically provisioned PersistentVolumes of this storage class are created with
            this reclaimPolicy. Defaults to Delete.
        volume_binding_mode(Text, optional): VolumeBindingMode indicates how PersistentVolumeClaims should be
            provisioned and bound. When unset, VolumeBindingImmediate is used. This field is only honored by servers
            that enable the VolumeScheduling feature.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.storage_k8s_io.v1.storage_class.present:
                - name: storage_class_1
                - allow_volume_expansion: true
                - allowed_topologies:
                    - match_label_expressions:
                        - key: "failure-domain.beta.kubernetes.io/zone"
                          values: ["us-west-2a"]
                - metadata:
                    name: "storage_class_1"
                - mount_options: ["soft"]
                - parameters:
                    type: "gp2"
                    iopsPerGB: "10"
                - provisioner: "kubernetes.io/aws-ebs"
                - reclaim_policy: "Retain"
                - volume_binding_mode: "WaitForFirstConsumer"
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Check for existing storage_class by name
    before = None
    if resource_id:
        storage_class = await hub.exec.k8s.client.StorageV1Api.read_storage_class(
            ctx, name=resource_id
        )
        if not storage_class["result"]:
            result["comment"] = storage_class["comment"]
            result["result"] = storage_class["result"]
            return result
        before = storage_class["ret"]

    # Update current state
    current_state = hub.tool.k8s.storage_k8s_io.v1.storage_class_utils.convert_raw_storage_class_to_present(
        storage_class=before
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "allow_volume_expansion": allow_volume_expansion,
        "metadata": metadata,
        "provisioner": provisioner,
        "reclaim_policy": reclaim_policy,
        "volume_binding_mode": volume_binding_mode,
    }

    if allowed_topologies:
        desired_state["allowed_topologies"] = allowed_topologies
    if mount_options:
        desired_state["mount_options"] = mount_options
    if parameters:
        desired_state["parameters"] = parameters

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )
    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1StorageClass"
    )
    if before:
        ret = await hub.exec.k8s.client.StorageV1Api.replace_storage_class(
            ctx, name=resource_id, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
        )
    else:
        ret = await hub.exec.k8s.client.StorageV1Api.create_storage_class(
            ctx, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
        )

    # Fetch the updated resource and update new_state
    storage_class = await hub.exec.k8s.client.StorageV1Api.read_storage_class(
        ctx, name=resource_id
    )
    if not storage_class["result"]:
        result["comment"] = result["comment"] + storage_class["comment"]
        result["result"] = storage_class["result"]
        return result
    after = storage_class["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.storage_k8s_io.v1.storage_class_utils.convert_raw_storage_class_to_present(
        storage_class=after
    )
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Delete a StorageClass.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.storage_k8s_io.v1.storage_class.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    if resource_id:
        storage_class = await hub.exec.k8s.client.StorageV1Api.read_storage_class(
            ctx, name=resource_id
        )
        if storage_class and storage_class["result"]:
            before = storage_class["ret"]
            result[
                "old_state"
            ] = hub.tool.k8s.storage_k8s_io.v1.storage_class_utils.convert_raw_storage_class_to_present(
                storage_class=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
        )
    else:
        ret = await hub.exec.k8s.client.StorageV1Api.delete_storage_class(
            ctx, name=resource_id
        )
        result["result"] = ret["result"]
        result["comment"] = (
            hub.tool.k8s.comment_utils.delete_comment(
                resource_type="k8s.storage_k8s_io.v1.storage_class", name=name
            )
            if result["result"]
            else ret["comment"]
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    list or watch objects of kind StorageClass


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.storage_k8s_io.v1.storage_class
    """

    ret = await hub.exec.k8s.client.StorageV1Api.list_storage_class(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe storage_class {ret['comment']}")
        return {}

    result = {}
    for storage_class in ret["ret"].items:
        storage_class_resource = hub.tool.k8s.storage_k8s_io.v1.storage_class_utils.convert_raw_storage_class_to_present(
            storage_class=storage_class
        )
        result[storage_class.metadata.name] = {
            "k8s.storage_k8s_io.v1.storage_class.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in storage_class_resource.items()
            ]
        }
    return result



import asyncio
import logging
from typing import Dict

import jmespath

logger = logging.getLogger(__name__)


async def wait(
    hub,
    ctx,
    waiter,
    waiter_config,
    initial_delay: int = None,
    err_graceful: bool = False,
):

    if not waiter or not waiter_config:
        raise NameError("Waiter or waiter config undefined")

    if initial_delay:
        await asyncio.sleep(initial_delay)

    method_handle = getattr(
        hub.exec.k8s.client, ".".join([waiter.api_class_name, waiter.operation])
    )
    delay = waiter_config.get("delay")
    max_attempts = waiter_config.get("max_attempts")
    retry = True
    while retry and max_attempts:
        ret = await method_handle(ctx, **waiter.arguments)
        if not ret["result"] and not err_graceful:
            raise RuntimeError(ret["comment"])

        status = ret["result"]
        comment = ret["comment"]
        data = ret["ret"].to_dict() if ret["result"] else {}

        if waiter.result_arguments:
            result = {
                key: extract(data, path)
                for key, path in waiter.result_arguments.items()
            }
        else:
            result = data

        retry = not bool(
            waiter.acceptor_function(status=status, data=result, comment=comment)
        )

        if retry:
            max_attempts -= 1
            await asyncio.sleep(delay)

    if not max_attempts:
        raise TimeoutError("Wait timed out.")


def extract(data: Dict, expr: str):
    return jmespath.search(expr, data)



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

apps/v1/DaemonSet
"""
import copy
from typing import Any
from typing import Dict
from typing import Tuple

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    spec: Dict,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Create a DaemonSet.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        spec(Dict): The desired behavior of this daemon set. More info:
            https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-
            status.
        timeout(Dict, optional): Timeout configuration for resource creation.
            * create (Dict) -- Timeout configuration for resource creation
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.apps.v1.daemon_set.present:
                - name: daemon_set_1
                - metadata: value
                - spec: value


            resource_is_present:
              k8s.apps.v1.daemon_set.present:
              - metadata:
                  name: nginx-ds
                  namespace: default
              - spec:
                  selector:
                    match_labels:
                      name: nginx
                  update_strategy:
                    rolling_update:
                      max_unavailable: 1
                    type: RollingUpdate
                  template:
                    metadata:
                      labels:
                        name: nginx
                    spec:
                      containers:
                      - image: nginx:1.14.2
                        image_pull_policy: IfNotPresent
                        name: nginx
                        resources:
                          limits:
                            memory: 200Mi
                          requests:
                            cpu: 100m
                            memory: 200Mi
                        termination_message_path: /dev/termination-log
                        termination_message_policy: File
                      restart_policy: Always
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"
    metadata["namespace"] = namespace

    # Check for existing daemon_set by name in namespace
    before = None
    if resource_id:
        daemon_set = await hub.exec.k8s.client.AppsV1Api.read_namespaced_daemon_set(
            ctx, name=resource_id, namespace=namespace
        )
        if not daemon_set["result"]:
            result["comment"] = daemon_set["comment"]
            result["result"] = daemon_set["result"]
            return result
        before = daemon_set["ret"]

    # Update current state
    current_state = (
        hub.tool.k8s.apps.v1.daemon_set_utils.convert_raw_daemon_set_to_present(
            daemon_set=before
        )
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    desired_state = {"resource_id": resource_id, "metadata": metadata, "spec": spec}

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.apps.v1.daemon_set", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.apps.v1.daemon_set", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1DaemonSet"
    )
    if before:
        ret = await hub.exec.k8s.client.AppsV1Api.replace_namespaced_daemon_set(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )
    else:
        ret = await hub.exec.k8s.client.AppsV1Api.create_namespaced_daemon_set(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )

    # Custom waiter for create/update
    waiter_config = hub.tool.k8s.waiter_utils.create_waiter_config(
        timeout_config=timeout.get("create") if timeout else None,
    )

    def create_acceptor(status: bool, data: Dict = None, comment: Tuple = None) -> bool:
        availability = data.get("availability") if data.get("availability") else None
        return True if availability else False

    arguments = {"name": resource_id, "namespace": namespace}
    result_arguments = {"availability": "status.number_available"}

    daemon_set_waiter = hub.tool.k8s.waiter_utils.create_waiter(
        api_class_name="AppsV1Api",
        operation="read_namespaced_daemon_set",
        arguments=arguments,
        result_arguments=result_arguments,
        acceptor_function=create_acceptor,
    )
    try:
        await hub.tool.k8s.custom_waiter.wait(
            ctx,
            waiter=daemon_set_waiter,
            waiter_config=waiter_config,
            err_graceful=False,
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
        return result

    # Fetch the updated resource and update new_state
    daemon_set = await hub.exec.k8s.client.AppsV1Api.read_namespaced_daemon_set(
        ctx, name=resource_id, namespace=namespace
    )
    if not daemon_set["result"]:
        result["comment"] = result["comment"] + daemon_set["comment"]
        result["result"] = daemon_set["result"]
        return result

    after = daemon_set["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.apps.v1.daemon_set_utils.convert_raw_daemon_set_to_present(
        daemon_set=after
    )
    return result


async def absent(
    hub, ctx, name: str, metadata: Dict, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Delete a DaemonSet.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        timeout(Dict, optional): Timeout configuration for resource deletion.
            * delete (Dict) -- Timeout configuration for resource deletion
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.apps.v1.daemon_set.absent:
                - name: value
                - metadata: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        daemon_set = await hub.exec.k8s.client.AppsV1Api.read_namespaced_daemon_set(
            ctx, name=resource_id, namespace=namespace
        )
        if daemon_set and daemon_set["result"]:
            before = daemon_set["ret"]
            result[
                "old_state"
            ] = hub.tool.k8s.apps.v1.daemon_set_utils.convert_raw_daemon_set_to_present(
                daemon_set=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )
    else:
        ret = await hub.exec.k8s.client.AppsV1Api.delete_namespaced_daemon_set(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        # Custom waiter for delete
        waiter_config = hub.tool.k8s.waiter_utils.create_waiter_config(
            timeout_config=timeout.get("delete") if timeout else None,
        )

        def delete_acceptor(
            status: bool, data: Dict = None, comment: Tuple = None
        ) -> bool:
            if not status and '"reason":"NotFound"' in comment[0]:
                return True
            if not status:
                raise RuntimeError(
                    f"Error while waiting for delete daemon_set : {comment[0]}"
                )
            return False

        arguments = {"name": resource_id, "namespace": namespace}
        result_arguments = {"availability": "status.number_available"}

        daemon_set_waiter = hub.tool.k8s.waiter_utils.create_waiter(
            api_class_name="AppsV1Api",
            operation="read_namespaced_daemon_set",
            arguments=arguments,
            result_arguments=result_arguments,
            acceptor_function=delete_acceptor,
        )
        try:
            await hub.tool.k8s.custom_waiter.wait(
                ctx,
                waiter=daemon_set_waiter,
                waiter_config=waiter_config,
                err_graceful=True,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.apps.v1.daemon_set", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.


    list or watch objects of kind DaemonSet.


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.apps.v1.daemon_set
    """

    ret = await hub.exec.k8s.client.AppsV1Api.list_daemon_set_for_all_namespaces(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe daemon_set {ret['comment']}")
        return {}

    result = {}
    for daemon_set in ret["ret"].items:
        daemon_set_resource = (
            hub.tool.k8s.apps.v1.daemon_set_utils.convert_raw_daemon_set_to_present(
                daemon_set=daemon_set
            )
        )
        result[daemon_set.metadata.name] = {
            "k8s.apps.v1.daemon_set.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in daemon_set_resource.items()
            ]
        }
    return result



"""State module for managing Kubernetes APIService."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: make_dataclass(
        "metadata",
        [
            ("name", str),
            ("namespace", str, field(default=None)),
            ("labels", Dict, field(default=None)),
            ("annotations", str, field(default=None)),
        ],
    ),
    spec: make_dataclass(
        "spec",
        [
            ("group_priority_minimum", int),
            ("group", str, field(default=None)),
            ("ca_bundle", str, field(default=None)),
            ("insecure_skip_tls_verify", bool, field(default=None)),
            ("version", str, field(default=None)),
            ("version_priority", int, field(default=None)),
            (
                "service",
                make_dataclass(
                    "service",
                    [
                        ("name", str, field(default=None)),
                        ("namespace", str, field(default=None)),
                        ("port", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates an APIService.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        metadata(dict):
            Standard object's metadata.
            More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata.

        spec(dict):
            Spec contains information for locating and communicating with a server.

    Request Syntax:

        .. code-block:: sls

            [api_service-name]:
              k8s.apiregistration_k8s_io.v1.api_service.present:
                - name: 'string'
                - metadata: Dict
                - spec: Dict
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.apiregistration_k8s_io.v1.api_service.present:
                - name: 'test-api-service'
                - metadata:
                    labels:
                      k8s-app: metrics-server
                    name: v1beta1.metrics.k8s.io
                - spec:
                    group: metrics.k8s.io
                    group_priority_minimum: 100
                    insecure_skip_tls_verify: true
                    service:
                      name: metrics-server
                      namespace: kube-system
                    version: v1beta1
                    version_priority: 100

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Check for existing api_service by name
    before = None
    if resource_id:
        api_service = await hub.exec.k8s.apiregistration_k8s_io.v1.api_service.get(
            ctx, name=name, resource_id=resource_id
        )
        if not api_service["result"] or not api_service["ret"]:
            result["result"] = False
            result["comment"] = tuple(api_service["comment"])
            return result
        before = api_service["ret"]

    # Update current state
    result["old_state"] = before

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "metadata": metadata,
        "spec": spec,
    }

    desired_state = hub.tool.k8s.state_utils.merge_arguments(
        desire_state=desired_state, current_state=result["old_state"]
    )

    is_change_detected = before is None or bool(
        differ.deep_diff(old=result["old_state"] or {}, new=desired_state)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=result["old_state"],
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1APIService"
    )
    if before:
        ret = await hub.exec.k8s.client.ApiregistrationV1Api.replace_api_service(
            ctx, name=resource_id, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )
    else:
        ret = await hub.exec.k8s.client.ApiregistrationV1Api.create_api_service(
            ctx, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )

    # Fetch the updated resource and update new_state
    api_service = await hub.exec.k8s.client.ApiregistrationV1Api.read_api_service(
        ctx, name=resource_id
    )
    if not api_service["result"]:
        result["comment"] = result["comment"] + api_service["comment"]
        result["result"] = api_service["result"]
        return result
    after = api_service["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.apiregistration_k8s_io.v1.api_service_utils.convert_raw_api_service_to_present(
        api_service=after
    )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes an APIService.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

    Request Syntax:

        .. code-block:: sls

            [api-service-name]:
              k8s.apiregistration_k8s_io.v1.api_service.absent:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.apiregistration_k8s_io.v1.api_service.absent:
                - name: "test-api-service"
                - resource_id: "test-api-service"
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    if resource_id:
        api_service = await hub.exec.k8s.apiregistration_k8s_io.v1.api_service.get(
            ctx, name=name, resource_id=resource_id
        )
        if api_service and api_service["result"]:
            before = api_service["ret"]
            result["old_state"] = before

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )
    else:
        ret = await hub.exec.k8s.client.ApiregistrationV1Api.delete_api_service(
            ctx, name=resource_id
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes the resource in a way that can be recreated/managed with the corresponding "present" function.

    list or watch objects of kind APIService.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.apiregistration_k8s_io.v1.api_service
    """

    result = {}
    ret = await hub.exec.k8s.apiregistration_k8s_io.v1.api_service.list(
        ctx, name="k8s.apiregistration_k8s_io.v1.api_service.describe"
    )
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe k8s.apiregistration_k8s_io.v1.api_service {ret['comment']}"
        )
        return {}

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "k8s.apiregistration_k8s_io.v1.api_service.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

core/v1/ServiceAccount
"""
import copy
from typing import Any
from typing import Dict
from typing import List

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    automount_service_account_token: bool = True,
    image_pull_secrets: List = None,
    secrets: List = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    create a ServiceAccount

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        automount_service_account_token(bool, optional): AutomountServiceAccountToken indicates whether pods running as this service account should have
            an API token automatically mounted. Can be overridden at the pod level.
        image_pull_secrets(List, optional): ImagePullSecrets is a list of references to secrets in the same namespace to use for pulling any
            images in pods that reference this ServiceAccount. ImagePullSecrets are distinct from Secrets
            because Secrets can be mounted in the pod, but ImagePullSecrets are only accessed by the
            kubelet. More info: https://kubernetes.io/docs/concepts/containers/images/#specifying-
            imagepullsecrets-on-a-pod.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        secrets(List, optional): Secrets is the list of secrets allowed to be used by pods running using this ServiceAccount.
            More info: https://kubernetes.io/docs/concepts/configuration/secret.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.core.v1.service_account.present:
                - name: value
                - metadata:
                  name: idem-service-account-test
                  namespace: default
                  labels:
                    example-label-name: example-label-value
                - automount_service_account_token: false
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"
    metadata["namespace"] = namespace

    # Check for existing service_account by name in namespace
    before = None
    if resource_id:
        service_account = (
            await hub.exec.k8s.client.CoreV1Api.read_namespaced_service_account(
                ctx, name=resource_id, namespace=namespace
            )
        )
        if not service_account["result"]:
            result["comment"] = service_account["comment"]
            result["result"] = service_account["result"]
            return result
        before = service_account["ret"]

    # Update current state
    current_state = hub.tool.k8s.core.v1.service_account_utils.convert_raw_service_account_to_present(
        service_account=before
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "automount_service_account_token": automount_service_account_token,
        "metadata": metadata,
    }

    if image_pull_secrets:
        desired_state["image_pull_secrets"] = image_pull_secrets

    # For the brownfield scenario, if ESM does not contain service_account information and secrets are None in the
    # SLS file, existing secrets of the service account should be used.
    if secrets:
        desired_state["secrets"] = secrets
    elif current_state:
        desired_state["secrets"] = current_state.get("secrets")

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )
    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.core.v1.service_account", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.core.v1.service_account", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1ServiceAccount"
    )
    if before:
        ret = await hub.exec.k8s.client.CoreV1Api.replace_namespaced_service_account(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.create_namespaced_service_account(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )

    # Fetch the updated resource and update new_state
    service_account = (
        await hub.exec.k8s.client.CoreV1Api.read_namespaced_service_account(
            ctx, name=resource_id, namespace=namespace
        )
    )
    if not service_account["result"]:
        result["comment"] = result["comment"] + service_account["comment"]
        result["result"] = service_account["result"]
        return result
    after = service_account["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.core.v1.service_account_utils.convert_raw_service_account_to_present(
        service_account=after
    )
    return result


async def absent(
    hub, ctx, name: str, metadata: Dict, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    delete a ServiceAccount

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.core.v1.service_account.absent:
                - name: value
                - metadata: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        service_account = (
            await hub.exec.k8s.client.CoreV1Api.read_namespaced_service_account(
                ctx, name=resource_id, namespace=namespace
            )
        )
        if service_account and service_account["result"]:
            before = service_account["ret"]
            result[
                "old_state"
            ] = hub.tool.k8s.core.v1.service_account_utils.convert_raw_service_account_to_present(
                service_account=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.delete_namespaced_service_account(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.core.v1.service_account", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    list or watch objects of kind ServiceAccount


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.core.v1.service_account
    """

    ret = await hub.exec.k8s.client.CoreV1Api.list_service_account_for_all_namespaces(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe service_account {ret['comment']}")
        return {}

    result = {}
    for service_account in ret["ret"].items:
        service_account_resource = hub.tool.k8s.core.v1.service_account_utils.convert_raw_service_account_to_present(
            service_account=service_account
        )
        result[service_account.metadata.name] = {
            "k8s.core.v1.service_account.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in service_account_resource.items()
            ]
        }
    return result



from typing import Any
from typing import Dict


def convert_raw_service_to_present(hub, service) -> Dict[str, Any]:
    if not service:
        return None

    result = {"resource_id": service.metadata.name}
    skip_attributes = [
        "api_version",
        "kind",
        "metadata.creation_timestamp",
        "metadata.generation",
        "metadata.managed_fields",
        "metadata.resource_version",
        "metadata.uid",
        "spec.progress_deadline_seconds",
        "spec.revision_history_limit",
        "spec.template.spec.dns_policy",
        "spec.template.spec.scheduler_name",
        "spec.template.spec.termination_grace_period_seconds",
        "status",
    ]
    result.update(
        hub.tool.k8s.marshaller.marshal(
            k8s_object=service,
            skip_attributes=skip_attributes,
            skip_empty_values=True,
        )
    )
    return result



"""Exec module for managing Kubernetes CoreV1 Service(s)."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    namespace: str = None,
) -> Dict[str, Any]:
    """Retrieves a Kubernetes CoreV1 Service.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The service.metadata.name of the Kubernetes CoreV1 Service.

        namespace(str, Optional):
            The Kubernetes namespace in which CoreV1 Service was created. Defaults to 'default' namespace in case None.

    Returns:
        Dict[str, Any]:
            Return a CoreV1 Service in a given namespace.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec k8s.core.v1.service.get name='service-name' resource_id='v1service' namespace='default'

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my-kubernetes-service:
              exec.run:
                - path: k8s.core.v1.service.get
                - kwargs:
                    name: 'service-name'
                    resource_id: 'v1service'
                    namespace: 'default'
    """
    result = dict(comment=[], ret=None, result=True)

    describe_ret = await hub.exec.k8s.client.CoreV1Api.read_namespaced_service(
        ctx, name=resource_id, namespace=namespace if namespace else "default"
    )
    if not describe_ret["result"]:
        # Do not return success=false when it is not found.
        if "ApiException" in str(describe_ret["comment"]) or "Reason: Not Found" in str(
            describe_ret["comment"]
        ):
            result["comment"].append(
                hub.tool.k8s.comment_utils.get_empty_comment(
                    resource_type="k8s.core.v1.service", name=resource_id
                )
            )
            result["comment"] += list(describe_ret["comment"])
            return result

        result["comment"] += list(describe_ret["comment"])
        result["result"] = False
        return result

    if not describe_ret["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.get_empty_comment(
                resource_type="k8s.core.v1.service", name=resource_id
            )
        )
        return result

    result["ret"] = hub.tool.k8s.core.v1.service_utils.convert_raw_service_to_present(
        service=describe_ret.get("ret"),
    )
    return result



from typing import Any
from typing import Dict

import pop.contract


def client(hub, path: str, context):
    """
    Get an AWS service client based on the path and call an operation on it
    idem will automatically populate credentials from acct in the client.

    The calls mirror the namespacing of boto3.client and have the same parameters

    If a response can be paginated, the full result will be asynchronously collected and returned.

    path::

        boto3.client.[service_name].[operation] [kwargs="values"]

    Examples:
        In these examples will use the service_name of "ec2" and operation of "create_vpc"

        Call from the CLI
        .. code-block: bash

            $ idem exec boto3.client.ec2.create_vpc CidrBlock="10.0.0.0/24"

        Call from code
        .. code-block: python

            await hub.exec.boto3.client.ec2.create_vpc(ctx, CidrBlock="10.0.0.0/24")

    :param hub:
    :param path: client.[service_name].[function_name]
    :param context: None
    :return: The result of the call
    """
    c, api_class, operation = path.split(".", maxsplit=2)
    assert c == "client"

    async def _client_caller(ctx, *args, **kwargs) -> Dict[str, Any]:
        result = {"comment": (), "ret": None, "result": True}
        try:
            ret: Dict[str, Any] = await hub.tool.k8s.client.exec(
                ctx, api_class, operation, *args, **kwargs
            )
            if hasattr(ret, "keys"):
                keys = sorted(ret.keys())
            else:
                keys = []
            result["comment"] = tuple(keys)
            result["ret"] = ret
            result["result"] = bool(ret)
        except Exception as e:
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
        return result

    return pop.contract.ContractedAsync(
        hub,
        contracts=[],
        func=_client_caller,
        ref=path,
        name=operation,
        implicit_hub=False,
    )



import json
from collections import OrderedDict


def get_plugins(hub, k8s_openapi_spec_str: str, resource: str):
    plugins = {}

    # Parse open-api spec to retrieve state file template metadata
    k8s_openapi_spec = json.loads(k8s_openapi_spec_str)
    resource_metadata = hub.pop_create.k8s.openapi_spec_parser.parse(
        spec=k8s_openapi_spec, resource_name=resource
    )
    if not resource_metadata:
        raise NameError(resource)

    # Populate plugin with resource metadata
    api_class = resource_metadata["create"].get("api_class")
    resource_name = hub.tool.format.case.snake(resource.split("/").pop())
    plugin_key = ".".join(
        [hub.tool.format.case.snake(item) for item in resource.split("/") if item]
    )

    create_function = resource_metadata["create"].get("function_name")
    delete_function = resource_metadata["delete"].get("function_name")
    update_function = resource_metadata["update"].get("function_name")
    get_function = resource_metadata["get"].get("function_name")
    list_function = resource_metadata["list"].get("function_name")

    create_params = OrderedDict()
    if "name" not in create_params:
        create_params["name"] = hub.pop_create.k8s.template.NAME_PARAMETER
    create_params["resource_id"] = hub.pop_create.k8s.template.RESOURCE_ID_PARAMETER
    create_params.update(resource_metadata["create"].get("params"))
    create_params["timeout"] = hub.pop_create.k8s.template.CREATE_TIMEOUT_PARAMETER

    state_params = (
        "{ "
        + ",".join(
            [
                f' "{hub.tool.format.case.snake(key)}": {hub.tool.format.case.snake(key)}'
                for key in create_params
            ]
        )
        + " }"
    )

    delete_params = OrderedDict()
    if "name" not in delete_params:
        delete_params["name"] = hub.pop_create.k8s.template.NAME_PARAMETER
    delete_params["resource_id"] = hub.pop_create.k8s.template.RESOURCE_ID_PARAMETER
    delete_params.update(resource_metadata["delete"].get("params"))
    delete_params["timeout"] = hub.pop_create.k8s.template.DELETE_TIMEOUT_PARAMETER

    shared_function_data = {
        "hardcoded": {
            "resource": resource_name,
            "payload_type": resource_metadata["create"].get("payload_type"),
            "create_function": f"hub.exec.k8s.client.{api_class}.{create_function}",
            "delete_function": f"hub.exec.k8s.client.{api_class}.{delete_function}",
            "update_function": f"hub.exec.k8s.client.{api_class}.{update_function}",
            "get_function": f"hub.exec.k8s.client.{api_class}.{get_function}",
            "list_function": f"hub.exec.k8s.client.{api_class}.{list_function}",
            "api_class": api_class,
            "waiter_function": get_function,
            "state_parameters": state_params,
        },
    }

    plugins[plugin_key] = {
        "doc": resource,
        "imports": [
            "import copy",
            "from typing import Any",
            "from typing import Dict",
            "from typing import List",
            "from typing import Text",
            "from typing import Tuple",
            "from deepdiff import DeepDiff",
            "__contracts__ = ['resource']",
        ],
        "functions": {
            "present": dict(
                doc=resource_metadata["create"].get("doc"),
                params=dict(name=create_params.pop("name"), **create_params),
                **shared_function_data,
            ),
            "absent": dict(
                doc=resource_metadata["delete"].get("doc"),
                params=dict(name=delete_params.pop("name"), **delete_params),
                **shared_function_data,
            ),
            "describe": dict(
                doc=resource_metadata["list"].get("doc"),
                params={},
                **shared_function_data,
            ),
        },
    }
    return plugins



"""Exec module for managing Kubernetes APIService."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Retrieves a Kubernetes APIService.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            The metadata.name of the Kubernetes APIService.

    Returns:
        Dict[str, Any]:
            Return a Kubernetes APIService.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec k8s.apiregistration_k8s_io.v1.api_service.get name='api-service-name' resource_id='api-service-1'

        Using in a state:

        .. code-block:: yaml

            my-kubernetes-api_service:
              exec.run:
                - path: k8s.apiregistration_k8s_io.v1.api_service.get
                - kwargs:
                    name: 'api-service-name'
                    resource_id: 'api-service-1'

    """
    result = dict(comment=[], ret=None, result=True)

    api_service = await hub.exec.k8s.client.ApiregistrationV1Api.read_api_service(
        ctx, name=resource_id
    )

    if not api_service["result"]:
        # Do not return success=false when it is not found.
        if "ApiException" in str(api_service["comment"]) or "Reason: Not Found" in str(
            api_service["comment"]
        ):
            result["comment"].append(
                hub.tool.k8s.comment_utils.get_empty_comment(
                    resource_type="k8s.apiregistration_k8s_io.v1.api_service",
                    name=resource_id,
                )
            )
            result["comment"] += list(api_service["comment"])
            return result

        result["comment"] += list(api_service["comment"])
        result["result"] = False
        return result

    if not api_service["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.get_empty_comment(
                resource_type="k8s.apiregistration_k8s_io.v1.api_service",
                name=resource_id,
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.k8s.apiregistration_k8s_io.v1.api_service_utils.convert_raw_api_service_to_present(
        api_service=api_service.get("ret"),
    )
    return result


async def list_(hub, ctx, name) -> Dict:
    """Retrieves list of Kubernetes ApiServices.

    Args:
        name(str, Optional): The name of the Idem state.

    Returns:
        Dict[bool, list, dict or None]:

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec k8s.apiregistration_k8s_io.v1.api_service.list name="idem_name"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: k8s.apiregistration_k8s_io.v1.api_service.list
                - kwargs:
                    name: my_resource

    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.k8s.client.ApiregistrationV1Api.list_api_service(ctx)

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.list_empty_comment(
                resource_type="k8s.apiregistration_k8s_io.v1.api_service", name=name
            )
        )
        return result
    for api_service in ret["ret"].items:
        converted_resource = hub.tool.k8s.apiregistration_k8s_io.v1.api_service_utils.convert_raw_api_service_to_present(
            api_service=api_service
        )
        result["ret"].append(converted_resource)
    return result



from typing import Any
from typing import Dict


def convert_raw_api_service_to_present(hub, api_service) -> Dict[str, Any]:
    if not api_service:
        return None

    result = {"resource_id": api_service.metadata.name}
    skip_attributes = [
        "api_version",
        "kind",
        "metadata.creation_timestamp",
        "metadata.generation",
        "metadata.managed_fields",
        "metadata.uid",
        "status",
    ]
    result.update(
        hub.tool.k8s.marshaller.marshal(
            k8s_object=api_service,
            skip_attributes=skip_attributes,
            skip_empty_values=True,
        )
    )
    return result



from typing import Any
from typing import Dict


def convert_raw_deployment_to_present(hub, deployment) -> Dict[str, Any]:
    if not deployment:
        return None

    result = {"resource_id": deployment.metadata.name}
    skip_attributes = [
        "api_version",
        "kind",
        "metadata.annotations.deployment.kubernetes.io/revision",
        "metadata.creation_timestamp",
        "metadata.generation",
        "metadata.managed_fields",
        "metadata.resource_version",
        "metadata.uid",
        "spec.progress_deadline_seconds",
        "spec.revision_history_limit",
        "spec.template.spec.dns_policy",
        "spec.template.spec.scheduler_name",
        "spec.template.spec.termination_grace_period_seconds",
        "status",
    ]
    result.update(
        hub.tool.k8s.marshaller.marshal(
            k8s_object=deployment,
            skip_attributes=skip_attributes,
            skip_empty_values=True,
        )
    )
    return result


def handle_default_values(hub, desired_state) -> Dict[str, Any]:
    if desired_state["spec"].get("strategy") is None:
        strategy = {
            "type": "RollingUpdate",
            "rolling_update": {"max_surge": "25%", "max_unavailable": "25%"},
        }
        desired_state["spec"]["strategy"] = strategy
    if desired_state["spec"].get("template"):
        pod_template = desired_state["spec"].get("template")
        if pod_template["spec"].get("restart_policy") is None:
            pod_template["spec"]["restart_policy"] = "Always"

        pod_containers = desired_state["spec"]["template"]["spec"].get("containers")
        for container in pod_containers:
            if container.get("image_pull_policy") is None:
                image = container.get("image")
                container["image_pull_policy"] = (
                    "Always" if ":latest" in image else "IfNotPresent"
                )

            if container.get("termination_message_policy") is None:
                container["termination_message_policy"] = "File"

            if container.get("termination_message_path") is None:
                container["termination_message_path"] = "/dev/termination-log"

    return desired_state



import os
from typing import Any
from typing import Dict

from kubernetes import client as k8s_client
from kubernetes.config import kube_config

ITERATION_FINISHED = object()

__func_alias__ = {"exec_": "exec"}


class K8sConfigurationError(Exception):
    def __init__(self, message="kube_config_path need to be set"):
        self.message = message
        super().__init__(self.message)


def load_kube_config(hub, ctx):
    config_file_path = os.environ.get("KUBE_CONFIG_PATH")
    context = os.environ.get("KUBE_CTX")

    k8s_profile = (hub.OPT.get("idem") or {}).get("acct_profile")

    if not config_file_path:
        config_file_path = (
            (((hub.OPT.get("acct") or {}).get("extras") or {}).get("k8s") or {}).get(
                k8s_profile
            )
            or {}
        ).get("kube_config_path")

    if not context:
        context = (
            (((hub.OPT.get("acct") or {}).get("extras") or {}).get("k8s") or {}).get(
                k8s_profile
            )
            or {}
        ).get("context")

    if not config_file_path:
        config_file_path = ctx.acct.get("kube_config_path")

    if not context:
        context = ctx.acct.get("context")

    if not config_file_path:
        raise K8sConfigurationError

    # If context is None, current_context from config file will be used.
    kube_config.load_kube_config(config_file=config_file_path, context=context)


async def exec_(
    hub,
    ctx,
    api_class: str,
    operation: str,
    *op_args,
    **op_kwargs: Dict[str, Any],
) -> Any:
    """
    :param hub:
    :param ctx:
    :param api_class: kubernetes API class name
    :param operation: The operation to run from the service client
    :param op_args: arguments to pass to the operation call
    :param op_kwargs: keyword arguments to pass to the operation call

    :return: The result of the operation call
    """
    hub.tool.k8s.client.load_kube_config(ctx)
    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in op_kwargs.items() if v is not None}
    api_instance = getattr(k8s_client, api_class)()
    hub.log.debug(f"Getting raw results for {api_instance}.{operation}")
    op = getattr(api_instance, operation)
    return await hub.pop.loop.wrap(op, *op_args, **kwargs)



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

v1/ConfigMap
"""
import copy
from typing import Any
from typing import Dict

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    immutable: bool = False,
    data: Dict = None,
    binary_data: Dict = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    create a ConfigMap

    Args:
        name(Text): An Idem name of the resource.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        binary_data(Dict, Optional): BinaryData contains the binary data. Each key must consist of alphanumeric characters, '-', '_'
            or '.'. BinaryData can contain byte sequences that are not in the UTF-8 range. The keys stored
            in BinaryData must not overlap with the ones in the Data field, this is enforced during
            validation process. Using this field will require 1.10+ apiserver and kubelet.
        data(Dict, Optional): Data contains the configuration data. Each key must consist of alphanumeric characters, '-', '_'
            or '.'. Values with non-UTF-8 byte sequences must use the BinaryData field. The keys stored in
            Data must not overlap with the keys in the BinaryData field, this is enforced during validation
            process.
        immutable(bool, Optional): Immutable, if set to true, ensures that data stored in the ConfigMap cannot be updated (only
            object metadata can be modified). If not set to true, the field can be modified at any time.
            Defaulted to nil.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.core.v1.config_map.present:
                - name: idem-config-map-test
                - data:
                      adminRoles: test-admin
                      clientType: test-clientType
                      viewerRoles: test-viewerRoles
                - metadata:
                      annotations:
                        meta.helm.sh/release-name: test-release
                        meta.helm.sh/release-namespace: test-release-default
                      labels:
                        app: test-app
                        environment: test-environment
                        product: test-product
                      name: idem-config-map-test
                      namespace: default
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"
    metadata["namespace"] = namespace

    # Check for existing config_map by name in namespace
    before = None
    if resource_id:
        config_map = await hub.exec.k8s.client.CoreV1Api.read_namespaced_config_map(
            ctx, name=resource_id, namespace=namespace
        )
        if not config_map["result"]:
            result["comment"] = config_map["comment"]
            result["result"] = config_map["result"]
            return result
        before = config_map["ret"]

    # Update current state
    current_state = (
        hub.tool.k8s.core.v1.config_map_utils.convert_raw_config_map_to_present(
            config_map=before
        )
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "metadata": metadata,
        "immutable": immutable,
    }

    if data:
        desired_state["data"] = data
    if binary_data:
        desired_state["binary_data"] = binary_data

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )
    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.core.v1.config_map", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.core.v1.config_map", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1ConfigMap"
    )
    if before:
        ret = await hub.exec.k8s.client.CoreV1Api.replace_namespaced_config_map(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.create_namespaced_config_map(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )

    # Fetch the updated resource and update new_state
    config_map = await hub.exec.k8s.client.CoreV1Api.read_namespaced_config_map(
        ctx, name=resource_id, namespace=namespace
    )
    if not config_map["result"]:
        result["comment"] = result["comment"] + config_map["comment"]
        result["result"] = config_map["result"]
        return result
    after = config_map["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.core.v1.config_map_utils.convert_raw_config_map_to_present(
        config_map=after
    )
    return result


async def absent(
    hub, ctx, name: str, metadata: Dict, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    delete a ConfigMap

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.core.v1.config_map.absent:
                - name: value
                - metadata: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        config_map = await hub.exec.k8s.client.CoreV1Api.read_namespaced_config_map(
            ctx, name=resource_id, namespace=namespace
        )
        if config_map and config_map["result"]:
            before = config_map["ret"]
            result[
                "old_state"
            ] = hub.tool.k8s.core.v1.config_map_utils.convert_raw_config_map_to_present(
                config_map=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.delete_namespaced_config_map(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.core.v1.config_map", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    list or watch objects of kind ConfigMap


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.core.v1.config_map
    """

    ret = await hub.exec.k8s.client.CoreV1Api.list_config_map_for_all_namespaces(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe config_map {ret['comment']}")
        return {}

    result = {}
    for config_map in ret["ret"].items:
        config_map_resource = (
            hub.tool.k8s.core.v1.config_map_utils.convert_raw_config_map_to_present(
                config_map=config_map
            )
        )
        result[config_map.metadata.name] = {
            "k8s.core.v1.config_map.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in config_map_resource.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

rbac.authorization.k8s.io/v1/ClusterRole
"""
import copy
from typing import Any
from typing import Dict
from typing import List

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    aggregation_rule: Dict = None,
    rules: List = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    create a ClusterRole

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        aggregation_rule(Dict, optional): AggregationRule is an optional field that describes how to build the Rules for this ClusterRole.
            If AggregationRule is set, then the Rules are controller managed and direct changes to Rules
            will be stomped by the controller.
        metadata(Dict): Standard object's metadata.
        rules(List. optional): Rules holds all the PolicyRules for this ClusterRole.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.rbac.v1.cluster_role.present:
              - name: value
              - metadata:
                  name: idem-test-cluster-role
              - rules:
                - api_groups:
                  - test.com
                  resources:
                  - vaconfigs
                  verbs:
                  - get
                  - list
                  - watch
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Check for existing cluster_role by name
    before = None
    if resource_id:
        cluster_role = (
            await hub.exec.k8s.client.RbacAuthorizationV1Api.read_cluster_role(
                ctx, name=resource_id
            )
        )
        if not cluster_role["result"]:
            result["comment"] = cluster_role["comment"]
            result["result"] = cluster_role["result"]
            return result
        before = cluster_role["ret"]

    # Update current state
    current_state = (
        hub.tool.k8s.rbac.v1.cluster_role_utils.convert_raw_cluster_role_to_present(
            cluster_role=before
        )
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "metadata": metadata,
    }

    if aggregation_rule:
        desired_state["aggregation_rule"] = aggregation_rule
    if rules:
        desired_state["rules"] = rules

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.rbac.v1.cluster_role", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.rbac.v1.cluster_role", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1ClusterRole"
    )
    if before:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.replace_cluster_role(
            ctx, name=resource_id, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )
    else:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.create_cluster_role(
            ctx, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )

    # Fetch the updated resource and update new_state
    cluster_role = await hub.exec.k8s.client.RbacAuthorizationV1Api.read_cluster_role(
        ctx, name=resource_id
    )
    if not cluster_role["result"]:
        result["comment"] = result["comment"] + cluster_role["comment"]
        result["result"] = cluster_role["result"]
        return result
    after = cluster_role["ret"]
    result[
        "new_state"
    ] = hub.tool.k8s.rbac.v1.cluster_role_utils.convert_raw_cluster_role_to_present(
        cluster_role=after
    )
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    delete a ClusterRole

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.rbac.v1.cluster_role.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = None
    if resource_id:
        cluster_role = (
            await hub.exec.k8s.client.RbacAuthorizationV1Api.read_cluster_role(
                ctx, name=resource_id
            )
        )
        if cluster_role and cluster_role["result"]:
            before = cluster_role["ret"]
            result[
                "old_state"
            ] = hub.tool.k8s.rbac.v1.cluster_role_utils.convert_raw_cluster_role_to_present(
                cluster_role=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )
    else:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.delete_cluster_role(
            ctx, name=resource_id
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.rbac.v1.cluster_role", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    list or watch objects of kind ClusterRole


    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.rbac.v1.cluster_role
    """

    ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.list_cluster_role(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe cluster_role {ret['comment']}")
        return {}

    result = {}
    for cluster_role in ret["ret"].items:
        cluster_role_resource = (
            hub.tool.k8s.rbac.v1.cluster_role_utils.convert_raw_cluster_role_to_present(
                cluster_role=cluster_role
            )
        )
        result[cluster_role.metadata.name] = {
            "k8s.rbac.v1.cluster_role.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in cluster_role_resource.items()
            ]
        }
    return result



from typing import Tuple


def create_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Created {resource_type} '{name}'",)


def would_create_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Would create {resource_type} '{name}'",)


def update_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Updated {resource_type} '{name}'",)


def get_empty_comment(hub, resource_type: str, name: str) -> str:
    return f"Get {resource_type} '{name}' result is empty"


def would_update_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Would update {resource_type} '{name}'",)


def delete_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Deleted {resource_type} '{name}'",)


def would_delete_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"Would delete {resource_type} '{name}'",)


def already_absent_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"{resource_type} '{name}' already absent",)


def already_exists_comment(hub, resource_type: str, name: str) -> Tuple:
    return (f"{resource_type} '{name}' already exists",)


def list_empty_comment(hub, resource_type: str, name: str) -> str:
    return f"List {resource_type} '{name}' result is empty"



"""State module for managening k8s RoleBinding."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: make_dataclass(
        "metadata",
        [
            ("name", str),
            ("namespace", str, field(default=None)),
            ("labels", Dict, field(default=None)),
            ("annotations", str, field(default=None)),
        ],
    ),
    role_ref: make_dataclass(
        "role_ref",
        [
            ("api_group", str),
            ("kind", str),
            ("name", str),
        ],
    ),
    subjects: List[
        make_dataclass(
            "subjects",
            [
                ("kind", str),
                ("name", str),
                ("namespace", str),
            ],
        )
    ],
    resource_id: str = None,
) -> Dict[str, Any]:
    """create a RoleBinding

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
        metadata(dict):
            Standard object's metadata.
        role_ref(dict):
            RoleRef can reference a Role in the current namespace or a ClusterRole in the global namespace.
            If the RoleRef cannot be resolved, the Authorizer must return an error.
        subjects(list):
            Subjects holds references to the objects the role applies to.

    Request Syntax:

        .. code-block:: sls

            [unmanaged_resource]:
              k8s.rbac.v1.role_binding.present:
                - name: 'string'
                - metadata: 'dict'
                - role_ref: 'dict'
                - subjects: 'dict'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.rbac.v1.role_binding.present:
                  - resource_id: system:controller:token-cleaner
                  - metadata:
                      annotations:
                        rbac.authorization.kubernetes.io/autoupdate: 'true'
                      labels:
                        kubernetes.io/bootstrapping: rbac-defaults
                      name: system:controller:token-cleaner
                      namespace: kube-system
                  - role_ref:
                      api_group: rbac.authorization.k8s.io
                      kind: Role
                      name: system:controller:token-cleaner
                  - subjects:
                    - kind: ServiceAccount
                      name: token-cleaner
                      namespace: kube-system"""

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace", "default")

    # Check for existing role_binding by name in namespace
    before = None
    if resource_id:
        role_binding = await hub.exec.k8s.rbac.v1.role_binding.get(
            ctx, name=name, resource_id=resource_id, namespace=namespace
        )
        if not role_binding["result"]:
            result["comment"] = role_binding["comment"]
            result["result"] = role_binding["result"]
            return result
        before = role_binding["ret"]

    # Update current state
    result["old_state"] = before

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "metadata": metadata,
        "role_ref": role_ref,
        "subjects": subjects,
    }

    is_change_detected = before is None or bool(
        differ.deep_diff(old=result["old_state"] or {}, new=desired_state)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=result["old_state"],
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.rbac.v1.role_binding", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.rbac.v1.role_binding", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1RoleBinding"
    )
    if before:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.replace_namespaced_role_binding(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
    else:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.create_namespaced_role_binding(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
    # Fetch the updated resource and update new_state
    role_binding = await hub.exec.k8s.rbac.v1.role_binding.get(
        ctx, name=name, resource_id=resource_id, namespace=namespace
    )
    if not role_binding["result"]:
        result["comment"] = result["comment"] + role_binding["comment"]
        result["result"] = role_binding["result"]
        return result
    result["new_state"] = role_binding["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    metadata: make_dataclass(
        "metadata",
        [
            ("name", str),
            ("namespace", str, field(default=None)),
            ("labels", Dict, field(default=None)),
            ("annotations", str, field(default=None)),
        ],
    ) = None,
    resource_id: str = None,
):
    """Delete a RoleBinding

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.
        metadata(dict, Optional):
            Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.

    Returns:
        Dict[str, Any]

    Request Syntax:

        .. code-block:: sls

            [k8s.rbac.v1.role_binding.name]:
              k8s.rbac.v1.role_binding.absent:
                - resource_id: 'string'
                - name: 'string'
                - metadata: 'dict'
    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.rbac.v1.role_binding.absent:
                - resource_id: 'k8s.rbac.v1.role_binding.id'
                - name: 'role_binding_name'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    namespace = "default"
    if metadata:
        namespace = metadata.get("namespace", "default")
    before = None
    if resource_id:
        role_binding = await hub.exec.k8s.rbac.v1.role_binding.get(
            ctx, name=name, resource_id=resource_id, namespace=namespace
        )
        if role_binding and role_binding["result"]:
            before = role_binding["ret"]
            result["old_state"] = before

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
    else:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.delete_namespaced_role_binding(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.rbac.v1.role_binding", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """list objects of kind RoleBinding

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.rbac.v1.role_binding
    """

    ret = await hub.exec.k8s.rbac.v1.role_binding.list(
        ctx, name="k8s.rbac.v1.role_binding.describe"
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe role_binding {ret['comment']}")
        return {}

    result = {}

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "k8s.rbac.v1.role_binding.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



"""Exec module for managing Kubernetes RoleBinding(s)."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str,
    namespace: str = "default",
) -> Dict[str, Any]:
    """Retrieves a Kubernetes RbacV1 RoleBinding.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            The metadata.name of the Kubernetes role_binding.

        namespace(str, Optional):
            The Kubernetes namespace in which role_bindingwas created.
            Defaults to 'default' namespace in case None.

    Returns:
        Dict[str, Any]:
            Return a role_binding in a given namespace.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec k8s.rbac.v1.role_binding.get name='role_binding-name' resource_id='role_binding' namespace='default'

        Using in a state:

        .. code-block:: yaml

            my-kubernetes-role-binding:
              exec.run:
                - path: k8s.rbac.v1.role_binding.get
                - kwargs:
                    name: 'role_binding-name'
                    resource_id: 'role_binding-name'
                    namespace: 'default'
    """
    result = dict(comment=[], ret=None, result=True)

    role_binding = (
        await hub.exec.k8s.client.RbacAuthorizationV1Api.read_namespaced_role_binding(
            ctx, name=resource_id, namespace=namespace
        )
    )

    if not role_binding["result"]:
        # Do not return success=false when it is not found.
        if "ApiException" in str(
            role_binding["comment"]
        ) and "Reason: Not Found" in str(role_binding["comment"]):
            result["comment"].append(
                hub.tool.k8s.comment_utils.get_empty_comment(
                    resource_type="k8s.rbac.v1.role_binding",
                    name=resource_id,
                )
            )
            result["comment"] += list(role_binding["comment"])
            return result

        result["comment"] += list(role_binding["comment"])
        result["result"] = False
        return result

    if not role_binding["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.get_empty_comment(
                resource_type="k8s.rbac.v1.role_binding", name=resource_id
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.k8s.rbac.v1.role_binding_utils.convert_raw_role_binding_to_present(
        role_binding=role_binding["ret"]
    )
    return result


async def list_(
    hub,
    ctx,
    name: str = None,
    namespace: str = None,
) -> Dict:
    """Retrieves list of Kubernetes RoleBindings.

    Args:
        name(str, Optional):
            The name of the Idem state.
        namespace(str, Optional):
            The Kubernetes namespace in which role_bindingwas created.
            Defaults to all namespace in case None.

    Returns:
        Dict[str, Any]:
            Returns role bindings in present format


    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec k8s.rbac.v1.role_binding.list name="idem_name" namespace="default"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: k8s.rbac.v1.role_binding.list
                - kwargs:
                    name: my_resource
                    namespace: default


    """
    result = dict(comment=[], ret=[], result=True)
    if namespace is None:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.list_role_binding_for_all_namespaces(
            ctx
        )
    else:
        ret = await hub.exec.k8s.client.RbacAuthorizationV1Api.list_namespaced_role_binding(
            ctx, namespace=namespace
        )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.list_empty_comment(
                resource_type="k8s.rbac.v1.role_binding", name=name
            )
        )
        return result
    for role_binding in ret["ret"].items:
        converted_resource = (
            hub.tool.k8s.rbac.v1.role_binding_utils.convert_raw_role_binding_to_present(
                role_binding=role_binding
            )
        )

        result["ret"].append(converted_resource)
    return result



"""Exec module for managing Kubernetes PersistentVolumeClaim(s)."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    namespace: str = "default",
) -> Dict[str, Any]:
    """Retrieves a Kubernetes CoreV1 PersistentVolumeClaim.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            The metadata.name of the Kubernetes CoreV1 PersistentVolumeClaim.

        namespace(str, Optional):
            The Kubernetes namespace in which CoreV1 PersistentVolumeClaim was created.
            Defaults to 'default' namespace in case None.

    Returns:
        Dict[str, Any]:
            Return a CoreV1 PersistentVolumeClaim in a given namespace.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec k8s.core.v1.persistent_volume_claim.get name='pvc-name' resource_id='pvc-12' namespace='default'

        Using in a state:

        .. code-block:: yaml

            my-kubernetes-pvc:
              exec.run:
                - path: k8s.core.v1.persistent_volume_claim.get
                - kwargs:
                    name: 'pvc-name'
                    resource_id: 'pvc-12'
                    namespace: 'default'
    """
    result = dict(comment=[], ret=None, result=True)

    persistent_volume_claim = (
        await hub.exec.k8s.client.CoreV1Api.read_namespaced_persistent_volume_claim(
            ctx, name=resource_id, namespace=namespace
        )
    )

    if not persistent_volume_claim["result"]:
        # Do not return success=false when it is not found.
        if "ApiException" in str(
            persistent_volume_claim["comment"]
        ) or "Reason: Not Found" in str(persistent_volume_claim["comment"]):
            result["comment"].append(
                hub.tool.k8s.comment_utils.get_empty_comment(
                    resource_type="k8s.core.v1.persistent_volume_claim",
                    name=resource_id,
                )
            )
            result["comment"] += list(persistent_volume_claim["comment"])
            return result

        result["comment"] += list(persistent_volume_claim["comment"])
        result["result"] = False
        return result

    if not persistent_volume_claim["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.get_empty_comment(
                resource_type="k8s.core.v1.persistent_volume_claim", name=resource_id
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.k8s.core.v1.persistent_volume_claim_utils.convert_raw_persistent_volume_claim_to_present(
        persistent_volume_claim=persistent_volume_claim.get("ret"),
    )
    return result


async def list_(hub, ctx, name) -> Dict:
    """Retrieves list of Kubernetes CoreV1 PersistentVolumeClaims.

    Args:
        name(str, Optional): The name of the Idem state.

    Returns:
        Dict[bool, list, dict or None]:

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec k8s.core.v1.persistent_volume_claim.list name="idem_name"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: k8s.core.v1.persistent_volume_claim.list
                - kwargs:
                    name: my_resource

    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.k8s.client.CoreV1Api.list_persistent_volume_claim_for_all_namespaces(
        ctx
    )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]:
        result["comment"].append(
            hub.tool.k8s.comment_utils.list_empty_comment(
                resource_type="k8s.core.v1.persistent_volume_claim", name=name
            )
        )
        return result
    for persistent_volume_claim in ret["ret"].items:
        converted_resource = hub.tool.k8s.core.v1.persistent_volume_claim_utils.convert_raw_persistent_volume_claim_to_present(
            persistent_volume_claim=persistent_volume_claim
        )
        result["ret"].append(converted_resource)
    return result



import re
from typing import Any
from typing import Dict
from typing import List

from kubernetes import client as k8s_client


def marshal(
    hub,
    k8s_object: Any,
    skip_attributes: List,
    skip_empty_values: bool,
    include_empty_string_keys: List = [],
):
    result = {}
    representation = k8s_object.to_dict()
    for key, value in representation.items():
        processed_value = process(
            key, value, skip_attributes, skip_empty_values, include_empty_string_keys
        )
        if skip_empty_values and not processed_value:
            continue
        result[key] = processed_value
    return result


def process(
    key: str,
    value: Any,
    skip_attributes: List,
    skip_empty_values: bool,
    include_empty_string_keys: List = [],
):
    # Handle skipping attributes
    if key in skip_attributes:
        return None

    # Handle skipping empty values
    if (not (key in include_empty_string_keys and value is not None)) and (
        (skip_empty_values and not value) or (skip_empty_values and value == "null")
    ):
        return None

    new_skip_attr = [
        item[len(key + ".") :] for item in skip_attributes if item.startswith(key + ".")
    ]
    if isinstance(value, list):
        result = []
        for item in value:
            processed_item = process(
                key, item, new_skip_attr, skip_empty_values, include_empty_string_keys
            )
            if processed_item or (
                key in include_empty_string_keys and processed_item is not None
            ):
                result.append(processed_item)
        return result
    if isinstance(value, dict):
        result = {}
        for k, v in value.items():
            processed_item = process(
                k, v, new_skip_attr, skip_empty_values, include_empty_string_keys
            )
            if processed_item or (
                key in include_empty_string_keys and processed_item is not None
            ):
                result[k] = processed_item
        return result
    return value


def unmarshal(hub, desired_state: Dict, k8s_model_name: str):
    k8s_class = getattr(k8s_client, k8s_model_name)
    arguments = {}
    for key, val in desired_state.items():
        if not k8s_class.openapi_types.get(key):
            continue
        result = val
        if isinstance(val, list):
            item_type = get_item_type(k8s_class, key)
            if item_type:
                result = [unmarshal(hub, item, item_type) for item in val]
        elif isinstance(val, dict):
            item_type = get_item_type(k8s_class, key)
            if item_type:
                result = unmarshal(hub, val, item_type)
        arguments[key] = result
    k8s_object = getattr(k8s_client, k8s_model_name)(**arguments)
    return k8s_object


def get_item_type(k8s_class, key: str):
    item_type = k8s_class.openapi_types.get(key).strip()
    if item_type.startswith("list") and re.search("list\\[(.+?)\\]", item_type):
        item_type = re.search("list\\[(.+?)\\]", item_type).group(1)
    if hasattr(k8s_client, item_type):
        return item_type
    return None



NAME_PARAMETER = {
    "default": None,
    "doc": "An Idem name of the resource.",
    "param_type": "Text",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}


RESOURCE_ID_PARAMETER = {
    "default": None,
    "doc": "An identifier of the resource in the provider",
    "param_type": "Text",
    "required": False,
    "target": "hardcoded",
    "target_type": "arg",
}

CREATE_TIMEOUT_PARAMETER = {
    "default": None,
    "doc": """Timeout configuration for resource creation.
            * create (Dict) -- Timeout configuration for resource creation
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40""",
    "param_type": "Dict",
    "required": False,
    "target": "hardcoded",
    "target_type": "arg",
}

DELETE_TIMEOUT_PARAMETER = {
    "default": None,
    "doc": """Timeout configuration for resource deletion.
            * delete (Dict) -- Timeout configuration for resource deletion
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40""",
    "param_type": "Dict",
    "required": False,
    "target": "hardcoded",
    "target_type": "arg",
}


PRESENT_REQUEST_FORMAT = r"""result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    # Check for existing {{ function.hardcoded.resource }} by name in namespace
    before = None
    if resource_id:
        {{ function.hardcoded.resource }} = await {{ function.hardcoded.get_function }}(
            ctx, name=resource_id, namespace=namespace
        )
        if not {{ function.hardcoded.resource }}["result"]:
            result["comment"] = {{ function.hardcoded.resource }}["comment"]
            result["result"] = {{ function.hardcoded.resource }}["result"]
            return result
        before = {{ function.hardcoded.resource }}["ret"]

    # Update current state
    # TODO: Implement hub.tool.{{ function.ref }}_utils.convert_raw_{{ function.hardcoded.resource }}_to_present function
    current_state = (
        hub.tool.{{ function.ref }}_utils.convert_raw_{{ function.hardcoded.resource }}_to_present(
            {{ function.hardcoded.resource }}=before
        )
    )
    result["old_state"] = current_state

    # Handle no change behaviour
    # TODO: The parameters of the desired_state should be the same as the present() input parameters.
    desired_state = {{ function.hardcoded.state_parameters }}

    is_change_detected = before is None or bool(
        DeepDiff(current_state, desired_state, ignore_order=True)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="{{ function.ref }}", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="{{ function.ref }}", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="{{ function.ref }}", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="{{ function.hardcoded.payload_type }}"
    )
    if before:
        ret = await {{ function.hardcoded.update_function }}(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
                resource_type="{{ function.ref }}", name=name
        )
    else:
        ret = await {{ function.hardcoded.create_function }}(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
                resource_type="{{ function.ref }}", name=name
        )

    # Custom waiter for create/update
    waiter_config = hub.tool.k8s.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get("create") if timeout else None,
    )

    def create_acceptor(status: bool, data: Dict = None, comment: Tuple = None) -> bool:
        # TODO implement wait acceptor method, return true when resource is created/updated
        return True

    arguments = {"name": resource_id, "namespace": namespace}

    # TODO add result arguments passed to create_acceptor
    # TODO format - key : jmes_path
    # TODO eg. result_arguments = { "status" : "status.condition" }
    result_arguments = {}

    {{ function.hardcoded.resource }}_waiter = hub.tool.k8s.waiter_utils.create_waiter(
        api_class_name="{{ function.hardcoded.api_class }}",
        operation="{{ function.hardcoded.waiter_function }}",
        arguments=arguments,
        result_arguments=result_arguments,
        acceptor_function=create_acceptor,
    )
    try:
        await hub.tool.k8s.custom_waiter.wait(
            ctx,
            waiter={{ function.hardcoded.resource }}_waiter,
            waiter_config=waiter_config,
            err_graceful=False
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
        return result

    # Fetch the updated resource and update new_state
    {{ function.hardcoded.resource }} = await {{ function.hardcoded.get_function }}(
        ctx, name=resource_id, namespace=namespace
    )
    if not {{ function.hardcoded.resource }}["result"]:
        result["comment"] = result["comment"] + {{ function.hardcoded.resource }}["comment"]
        result["result"] = {{ function.hardcoded.resource }}["result"]
        return result
    after = {{ function.hardcoded.resource }}["ret"]
    result[
        "new_state"
    ] = hub.tool.{{ function.ref }}_utils.convert_raw_{{ function.hardcoded.resource }}_to_present(
        {{ function.hardcoded.resource }}=after
    )
    return result"""

ABSENT_REQUEST_FORMAT = r"""result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        {{ function.hardcoded.resource }} = await {{ function.hardcoded.get_function }}(
            ctx, name=resource_id, namespace=namespace
        )
        if {{ function.hardcoded.resource }} and {{ function.hardcoded.resource }}["result"]:
            before = {{ function.hardcoded.resource }}["ret"]
            result[
                "old_state"
            ] = hub.tool.{{ function.ref }}_utils.convert_raw_{{ function.hardcoded.resource }}_to_present(
                {{ function.hardcoded.resource }}=before
            )

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="{{ function.ref }}", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="{{ function.ref }}", name=name
        )
    else:
        ret = await {{ function.hardcoded.delete_function }}(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        # Custom waiter for delete
        waiter_config = hub.tool.k8s.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )

        def delete_acceptor(status: bool, data: Dict = None, comment: Tuple = None) -> bool:
            # TODO implement wait acceptor method, return true when resource is deleted
            return True

        arguments = {"name": name, "namespace": namespace}

        # TODO add result arguments passed to delete_acceptor
        # TODO format - key : jmes_path
        # TODO eg. result_arguments = { "status" : "status.condition" }
        result_arguments = {}

        {{ function.hardcoded.resource }}_waiter = hub.tool.k8s.waiter_utils.create_waiter(
            api_class_name="{{ function.hardcoded.api_class }}",
            operation="{{ function.hardcoded.waiter_function }}",
            arguments=arguments,
            result_arguments=result_arguments,
            acceptor_function=delete_acceptor,
        )
        try:
            await hub.tool.k8s.custom_waiter.wait(
                ctx,
                waiter={{ function.hardcoded.resource }}_waiter,
                waiter_config=waiter_config,
                err_graceful=True
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
                resource_type="{{ function.ref }}", name=name
        )
    return result"""

DESCRIBE_REQUEST_FORMAT = r"""ret = await {{ function.hardcoded.list_function }}(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe {{ function.hardcoded.resource }} {ret['comment']}")
        return {}

    result = {}
    for {{ function.hardcoded.resource }} in ret["ret"].items:
        {{ function.hardcoded.resource }}_resource = (
            hub.tool.{{ function.ref }}_utils.convert_raw_{{ function.hardcoded.resource }}_to_present(
                {{ function.hardcoded.resource }}={{ function.hardcoded.resource }}
            )
        )
        result[{{ function.hardcoded.resource }}.metadata.name] = {
            "{{ function.ref }}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in {{ function.hardcoded.resource }}_resource.items()
            ]
        }
    return result"""



def translate_type(hub, type_name: str):
    if type_name == "string":
        return "Text"
    elif type_name in ("map", "object", "structure"):
        return "Dict"
    elif type_name in ("list", "array"):
        return "List"
    elif type_name == "boolean":
        return "bool"
    elif type_name in ("integer", "long"):
        return "int"
    elif type_name in ("float", "double"):
        return "float"
    elif type_name == "timestamp":
        return "Text"
    elif type_name == "blob":
        return "ByteString"
    elif type_name == "object":
        return "Dict"
    else:
        raise NameError(type_name)



"""State module for managing Kubernetes PersistentVolumeClaim(s)."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: make_dataclass(
        "metadata",
        [
            ("name", str),
            ("namespace", str, field(default=None)),
            ("labels", Dict, field(default=None)),
            ("annotations", str, field(default=None)),
        ],
    ),
    spec: make_dataclass(
        "spec",
        [
            ("access_modes", List[str], field(default=None)),
            ("volume_mode", str, field(default=None)),
            ("storage_class_name", str, field(default=None)),
            (
                "resources",
                make_dataclass(
                    "resources",
                    [
                        (
                            "request",
                            make_dataclass(
                                "request",
                                [("storage", str, field(default=None))],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "selector",
                make_dataclass(
                    "selector",
                    [
                        ("match_labels", Dict, field(default=None)),
                        ("match_expressions", List[Dict], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates a PersistentVolumeClaim.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        metadata(dict):
            Standard object's metadata.
            More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata.

        spec(dict):
            Spec defines the desired characteristics of a volume requested by a pod author.
            More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#persistentvolumeclaims.

    Request Syntax:

        .. code-block:: sls

            [pvc-name]:
              k8s.core.v1.persistent_volume_claim.present:
                - name: 'string'
                - metadata: Dict
                - spec: Dict
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.core.v1.persistent_volume_claim.present:
                - name: "pvc-1"
                - metadata:
                    name: "pvc-1"
                    labels:
                      name: "pvc-1"
                    annotations:
                      volume.beta.kubernetes.io/storage-class: "sc-1"
                - spec:
                    access_modes:
                      - ReadWriteMany
                    resources:
                      requests:
                        storage: 1Mi

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"
    metadata["namespace"] = namespace

    # Check for existing persistent_volume_claim by name in namespace
    before = None
    if resource_id:
        persistent_volume_claim = (
            await hub.exec.k8s.core.v1.persistent_volume_claim.get(
                ctx, name=name, resource_id=resource_id, namespace=namespace
            )
        )
        if not persistent_volume_claim["result"] or not persistent_volume_claim["ret"]:
            result["result"] = False
            result["comment"] = tuple(persistent_volume_claim["comment"])
            return result
        before = persistent_volume_claim["ret"]

    # Update current state
    result["old_state"] = before

    # Handle no change behaviour
    desired_state = {
        "resource_id": resource_id,
        "metadata": metadata,
        "spec": spec,
    }

    desired_state = hub.tool.k8s.state_utils.merge_arguments(
        desire_state=desired_state, current_state=result["old_state"]
    )
    is_change_detected = before is None or bool(
        differ.deep_diff(old=result["old_state"] or {}, new=desired_state)
    )

    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=result["old_state"],
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.core.v1.persistent_volume_claim", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.core.v1.persistent_volume_claim", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1PersistentVolumeClaim"
    )
    if before:
        ret = await hub.exec.k8s.client.CoreV1Api.replace_namespaced_persistent_volume_claim(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.create_namespaced_persistent_volume_claim(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )

    # Fetch the updated resource and update new_state
    persistent_volume_claim = await hub.exec.k8s.core.v1.persistent_volume_claim.get(
        ctx, name=name, resource_id=resource_id, namespace=namespace
    )
    if not persistent_volume_claim["result"]:
        result["comment"] += persistent_volume_claim["comment"]
        result["result"] = persistent_volume_claim["result"]
        return result

    result["new_state"] = persistent_volume_claim["ret"]
    return result


async def absent(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes a PersistentVolumeClaim.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        metadata(dict):
            Standard object's metadata.
            More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata.

    Request Syntax:

        .. code-block:: sls

            [service-name]:
              k8s.core.v1.service.absent:
                - name: 'string'
                - metadata: Dict
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.core.v1.persistent_volume_claim.absent:
                - name: "pvc-1"
                - resource_id: "pvc-1"
                - metadata:
                    namespace: "default"
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        persistent_volume_claim = (
            await hub.exec.k8s.core.v1.persistent_volume_claim.get(
                ctx, name=name, resource_id=resource_id, namespace=namespace
            )
        )
        if persistent_volume_claim and persistent_volume_claim["result"]:
            before = persistent_volume_claim["ret"]
            result["old_state"] = before

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.delete_namespaced_persistent_volume_claim(
            ctx, name=resource_id, namespace=namespace
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
            resource_type="k8s.core.v1.persistent_volume_claim", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes the resource in a way that can be recreated/managed with the corresponding "present" function.

    list or watch objects of kind PersistentVolumeClaim.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.core.v1.persistent_volume_claim

    """
    result = {}
    ret = await hub.exec.k8s.core.v1.persistent_volume_claim.list(
        ctx, name="k8s.core.v1.persistent_volume_claim.describe"
    )
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe k8s.core.v1.persistent_volume_claim {ret['comment']}"
        )
        return {}

    for resource in ret["ret"]:
        result[resource.get("resource_id")] = {
            "k8s.core.v1.persistent_volume_claim.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

core/v1/Service
"""
import copy
from typing import Any
from typing import Dict
from typing import Tuple

from deepdiff import DeepDiff

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    metadata: Dict,
    spec: Dict,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""create a Service

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.
        spec(Dict): Spec defines the behavior of a service. https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#spec-and-status.
        timeout(Dict, optional): Timeout configuration for resource creation.
            * create (Dict) -- Timeout configuration for resource creation
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40. Defaults to None.

    Request Syntax:

        .. code-block:: sls

            [service-name]:
              k8s.core.v1.service.present:
                - name: "string"
                - metadata: "Dict"
                - spec: "Dict"
                - resource_id: "string"
                - timeout: "Dict"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              k8s.core.v1.service.present:
                - name: idem-test-service
                - metadata:
                    name: idem-test-secret
                    namespace: default
                - spec:
                    type: NodePort
                    selector:
                      app: echo-hostname
                    ports:
                      - nodePort: 30163
                        port: 8080
                        targetPort: 80
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Get namespace
    namespace = metadata.get("namespace") if "namespace" in metadata else "default"
    metadata["namespace"] = namespace

    # Check for existing service by name in namespace
    before = None
    if resource_id:
        service = await hub.exec.k8s.core.v1.service.get(
            ctx, name=name, resource_id=resource_id, namespace=namespace
        )
        if not service["result"] or not service["ret"]:
            result["result"] = False
            result["comment"] = tuple(service["comment"])
            return result
        before = service["ret"]

    # Update current state
    result["old_state"] = before

    # Handle no change behaviour
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "metadata": metadata,
        "spec": spec,
        "timeout": timeout,
    }
    is_change_detected = before is None or bool(
        DeepDiff(result["old_state"], desired_state, ignore_order=True)
    )
    if not is_change_detected:
        result["comment"] = hub.tool.k8s.comment_utils.already_exists_comment(
            resource_type="k8s.core.v1.service", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.k8s.test_state_utils.generate_test_state(
            enforced_state=result["old_state"],
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.k8s.comment_utils.would_update_comment(
                resource_type="k8s.core.v1.service", name=name
            )
            if before
            else hub.tool.k8s.comment_utils.would_create_comment(
                resource_type="k8s.core.v1.service", name=name
            )
        )
        return result

    # Handle actual resource create or update
    body = hub.tool.k8s.marshaller.unmarshal(
        desired_state=desired_state, k8s_model_name="V1Service"
    )
    if before:
        ret = await hub.exec.k8s.client.CoreV1Api.replace_namespaced_service(
            ctx, name=resource_id, namespace=namespace, body=body
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.update_comment(
            resource_type="k8s.core.v1.service", name=name
        )
    else:
        ret = await hub.exec.k8s.client.CoreV1Api.create_namespaced_service(
            ctx, namespace=namespace, body=body
        )
        resource_id = body.metadata.name
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.k8s.comment_utils.create_comment(
            resource_type="k8s.core.v1.service", name=name
        )

    # Custom waiter for create/update
    waiter_config = hub.tool.k8s.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get("create") if timeout else None,
    )

    def create_acceptor(status: bool, data: Dict = None, comment: Tuple = None) -> bool:
        service_type = data.get("type")
        if service_type == "LoadBalancer":
            load_balancer = (
                data.get("load_balancer") if data.get("load_balancer") else {}
            )
            if load_balancer:
                ingress = load_balancer.get("ingress") or []
                if ingress and (not ingress[0].get("hostname")):
                    hub.log.debug(
                        f"'status.load_balancer.ingress[0].hostname' is still empty for service: "
                        f"{data.get('resource_id')}"
                    )
                    return False
        return True

    arguments = {"name": resource_id, "namespace": namespace}
    result_arguments = {
        "load_balancer": "status.load_balancer",
        "type": "spec.type",
        "resource_id": "metadata.name",
    }
    service_waiter = hub.tool.k8s.waiter_utils.create_waiter(
        api_class_name="CoreV1Api",
        operation="read_namespaced_service",
        arguments=arguments,
        result_arguments=result_arguments,
        acceptor_function=create_acceptor,
    )
    try:
        await hub.tool.k8s.custom_waiter.wait(
            ctx, waiter=service_waiter, waiter_config=waiter_config, err_graceful=False
        )
    except Exception as e:
        result["comment"] += (str(e),)
        result["result"] = False
        return result

    # Fetch the updated resource and update new_state
    service = await hub.exec.k8s.core.v1.service.get(
        ctx, name=name, resource_id=resource_id, namespace=namespace
    )
    if not service["result"]:
        result["comment"] += service["comment"]
        result["result"] = service["result"]
        return result
    after = service["ret"]
    result["new_state"] = after
    return result


async def absent(
    hub, ctx, name: str, metadata: Dict, resource_id: str = None
) -> Dict[str, Any]:
    """delete a Service

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): An identifier of the resource in the provider. Defaults to None.
        metadata(Dict): Standard object's metadata. More info: https://git.k8s.io/community/contributors/devel/sig-
            architecture/api-conventions.md#metadata.

    Request Syntax:
      .. code-block:: sls
           [service-name]:
              k8s.core.v1.service.absent:
                - name: 'string'
                - metadata: 'Dict'
                - resource_id: 'string'

    Returns:
      Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              k8s.core.v1.service.absent:
                - name: value
                - metadata: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    namespace = metadata.get("namespace") if "namespace" in metadata else "default"

    before = None
    if resource_id:
        service = await hub.exec.k8s.core.v1.service.get(
            ctx, name=name, resource_id=resource_id, namespace=namespace
        )
        if service and service["result"]:
            before = service["ret"]
            result["old_state"] = before

    if not before:
        result["comment"] = hub.tool.k8s.comment_utils.already_absent_comment(
            resource_type="k8s.core.v1.service", name=name
        )
    else:
        if ctx.get("test", False):
            result["comment"] = hub.tool.k8s.comment_utils.would_delete_comment(
                resource_type="k8s.core.v1.service", name=name
            )
            return result
        else:
            ret = await hub.exec.k8s.client.CoreV1Api.delete_namespaced_service(
                ctx, name=resource_id, namespace=namespace
            )
            if not ret["result"]:
                result["result"] = False
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.k8s.comment_utils.delete_comment(
                resource_type="k8s.core.v1.service", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    list or watch objects of kind Service

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe k8s.core.v1.service
    """

    ret = await hub.exec.k8s.client.CoreV1Api.list_service_for_all_namespaces(
        ctx,
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe service {ret['comment']}")
        return {}

    result = {}
    for service in ret["ret"].items:
        service_resource = (
            hub.tool.k8s.core.v1.service_utils.convert_raw_service_to_present(
                service=service
            )
        )
        result[service.metadata.name] = {
            "k8s.core.v1.service.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in service_resource.items()
            ]
        }
    return result
